package ru.dbotthepony.mc.otm.android;

import com.mojang.blaze3d.vertex.PoseStack;
import net.minecraft.ChatFormatting;
import net.minecraft.client.Minecraft;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.network.chat.Component;
import net.minecraft.network.chat.TranslatableComponent;
import net.minecraft.resources.ResourceLocation;
import net.minecraft.world.item.Item;
import net.minecraft.world.item.ItemStack;
import net.minecraft.world.item.Items;
import net.minecraftforge.common.util.INBTSerializable;
import net.minecraftforge.registries.ForgeRegistry;
import net.minecraftforge.registries.RegistryManager;
import ru.dbotthepony.mc.otm.Registry;
import ru.dbotthepony.mc.otm.capability.AndroidCapabilityPlayer;
import ru.dbotthepony.mc.otm.capability.MatteryCapability;
import ru.dbotthepony.mc.otm.network.MatteryNetworking;
import ru.dbotthepony.mc.otm.network.android.AndroidResearchRequestPacket;
import ru.dbotthepony.mc.otm.screen.SkinElement;

import javax.annotation.Nullable;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.function.Supplier;

public abstract class AndroidResearch implements INBTSerializable<CompoundTag> {
	public final AndroidResearchType<?> type;
	public final AndroidCapabilityPlayer capability;

	protected boolean researched = false;
	public boolean dirty = true;

	public void markDirty() {
		dirty = true;
	}

	public boolean isResearched() {
		return researched;
	}

	public AndroidResearch(AndroidResearchType<?> type, AndroidCapabilityPlayer capability) {
		this.capability = capability;
		this.type = type;
	}

	abstract public List<AndroidResearchType<?>> getPrerequisites();
	abstract public void onResearched();
	abstract public void consumeCost();
	abstract public boolean canAfford();

	@Nullable
	public SkinElement getIcon() {
		return null;
	}

	public Component getDisplayName() {
		return type.getDisplayName();
	}

	public List<Component> getTooltip() {
		return List.of(getDisplayName());
	}

	public boolean canResearch() {
		for (var preq : getPrerequisites()) {
			if (!capability.getResearch(preq).isResearched()) {
				return false;
			}
		}

		return canAfford();
	}

	public boolean research(boolean force) {
		if (researched)
			return false;

		if (!force && !canResearch())
			return false;

		consumeCost();
		researched = true;
		onResearched();

		markDirty();

		return true;
	}

	@Nullable
	public Component getIconText() {
		return null;
	}

	public boolean research() {
		return research(false);
	}

	public void researchClient() {
		if (!researched && canAfford() && capability.isAndroid())
			MatteryNetworking.CHANNEL.sendToServer(new AndroidResearchRequestPacket(type));
	}

	@Override
	public CompoundTag serializeNBT() {
		var nbt = new CompoundTag();
		serializeNBT(nbt);
		return nbt;
	}

	public void serializeNBT(CompoundTag tag) {
		tag.putByte("researched", researched ? (byte) 1 : 0);
	}

	public void deserializeNBT(CompoundTag nbt) {
		researched = nbt.getByte("researched") > 0;
	}

	public static class Builder {
		record DeferredItemStack(ResourceLocation id, int amount) {}
		record DeferredFeature(ResourceLocation id, int level) {
			DeferredFeature(ResourceLocation id, int level) {
				this.id = id;
				this.level = level;
			}

			DeferredFeature(ResourceLocation id) {
				this(id, 0);
			}
		}

		public int experience = 0;
		public Component text;
		public boolean has_description = false;
		public Component custom_description;
		public Component custom_name;
		public final ArrayList<DeferredItemStack> items = new ArrayList<>();
		public final ArrayList<ResourceLocation> prerequisites = new ArrayList<>();
		public final ArrayList<DeferredFeature> feature_results = new ArrayList<>();
		public SkinElement icon;

		public Builder addPrerequisite(ResourceLocation location) {
			prerequisites.add(location);
			return this;
		}

		public Builder addPrerequisite(AndroidResearchType<?> location) {
			prerequisites.add(Objects.requireNonNull(location.getRegistryName()));
			return this;
		}

		public Builder addFeatureResult(ResourceLocation location) {
			feature_results.add(new DeferredFeature(location));
			return this;
		}

		public Builder addFeatureResult(AndroidFeatureType<?> location) {
			feature_results.add(new DeferredFeature(Objects.requireNonNull(location.getRegistryName())));
			return this;
		}

		public Builder addFeatureResult(ResourceLocation location, int level) {
			feature_results.add(new DeferredFeature(location, level));
			return this;
		}

		public Builder addFeatureResult(AndroidFeatureType<?> location, int level) {
			feature_results.add(new DeferredFeature(Objects.requireNonNull(location.getRegistryName()), level));
			return this;
		}

		public Builder addItem(ResourceLocation location) {
			items.add(new DeferredItemStack(location, 1));
			return this;
		}

		public Builder addItem(ResourceLocation location, int amount) {
			items.add(new DeferredItemStack(location, amount));
			return this;
		}

		public Builder setExperienceCost(int experience) {
			this.experience = experience;
			return this;
		}

		public Builder withDescription() {
			this.has_description = true;
			return this;
		}

		public Builder withDescription(Component text) {
			this.has_description = true;
			this.custom_description = text;
			return this;
		}

		public Builder withName(Component text) {
			this.custom_name = text;
			return this;
		}

		public Builder withIcon(SkinElement icon) {
			this.icon = icon;
			return this;
		}

		public Builder withIconText(Component text) {
			this.text = text;
			return this;
		}

		private final ArrayList<ItemStack> resolved_stacks = new ArrayList<>();
		private final ArrayList<AndroidResearchType<?>> resolved_preq = new ArrayList<>();

		record ResolvedFeature(AndroidFeatureType<?> type, int level) {}

		private final ArrayList<ResolvedFeature> resolved_features = new ArrayList<>();

		private boolean resolved = false;

		private void resolve() {
			if (!resolved) {
				resolved = true;

				for (var item : items) {
					var get_item = RegistryManager.ACTIVE.getRegistry(Item.class).getValue(item.id);

					if (get_item != null && get_item != Items.AIR) {
						resolved_stacks.add(new ItemStack(get_item, item.amount));
					}
				}

				for (var entry : prerequisites) {
					var get = Registry.ANDROID_RESEARCH().getValue(entry);

					if (get != null) {
						resolved_preq.add(get);
					} else {
						throw new IllegalArgumentException("Can not find android research " + entry);
					}
				}

				for (var entry : feature_results) {
					var get = Registry.ANDROID_FEATURES().getValue(entry.id);

					if (get != null) {
						resolved_features.add(new ResolvedFeature(get, entry.level));
					} else {
						throw new IllegalArgumentException("Can not find android feature " + entry);
					}
				}
			}
		}

		public AndroidResearchType.AndroidResearchFactory<AndroidResearch> build() {
			return (type, capability) -> {
				resolve();
				final var list_cp = List.copyOf(resolved_preq);

				return new AndroidResearch(type, capability) {
					@Override
					public Component getDisplayName() {
						if (custom_name != null)
							return custom_name;

						return super.getDisplayName();
					}

					@Nullable
					@Override
					public SkinElement getIcon() {
						return icon;
					}

					@Nullable
					@Override
					public Component getIconText() {
						return text;
					}

					@Override
					public List<Component> getTooltip() {
						var list = new ArrayList<>(super.getTooltip());

						if (custom_description != null) {
							list.add(custom_description);
						} else if (has_description) {
							list.add(new TranslatableComponent("android_research." + type.getRegistryName().getNamespace() + "." + type.getRegistryName().getPath() + ".description"));
						}

						if (!researched) {
							if (experience > 0) {
								if (capability.ply.experienceLevel >= experience) {
									list.add(new TranslatableComponent("otm.android_station.research.xp_cost", experience).withStyle(ChatFormatting.DARK_GREEN));
								} else {
									list.add(new TranslatableComponent("otm.android_station.research.xp_cost", experience).withStyle(ChatFormatting.DARK_RED));
								}
							}

							for (var stack : resolved_stacks) {
								boolean hit = false;

								if (stack.getTag() == null) {
									for (var inv_stack : capability.ply.getInventory().items) {
										if (inv_stack.is(stack.getItem()) && inv_stack.getCount() >= stack.getCount()) {
											hit = true;
											break;
										}
									}
								} else {
									for (var inv_stack : capability.ply.getInventory().items) {
										if (inv_stack.is(stack.getItem()) && inv_stack.getCount() >= stack.getCount() && inv_stack.getTag() != null && inv_stack.getTag().equals(stack.getTag())) {
											hit = true;
											break;
										}
									}
								}

								if (hit) {
									list.add(new TranslatableComponent("otm.android_station.research.item", stack.getDisplayName(), stack.getCount()).withStyle(ChatFormatting.DARK_GREEN));
								} else {
									list.add(new TranslatableComponent("otm.android_station.research.item", stack.getDisplayName(), stack.getCount()).withStyle(ChatFormatting.DARK_RED));
								}
							}
						}

						return list;
					}

					@Override
					public List<AndroidResearchType<?>> getPrerequisites() {
						return list_cp;
					}

					@Override
					public void onResearched() {
						for (var feature : resolved_features) {
							var get = capability.getFeature(feature.type);

							if (get == null) {
								get = capability.addFeature(feature.type);
								get.setLevel(feature.level);
							} else if (get.getLevel() < feature.level) {
								get.setLevel(feature.level);
							}
						}
					}

					@Override
					public boolean canAfford() {
						if (capability.ply.experienceLevel < experience)
							return false;

						for (var stack : resolved_stacks) {
							boolean hit = false;

							if (stack.getTag() == null) {
								for (var inv_stack : capability.ply.getInventory().items) {
									if (inv_stack.is(stack.getItem()) && inv_stack.getCount() >= stack.getCount()) {
										hit = true;
										break;
									}
								}
							} else {
								for (var inv_stack : capability.ply.getInventory().items) {
									if (inv_stack.is(stack.getItem()) && inv_stack.getCount() >= stack.getCount() && inv_stack.getTag() != null && inv_stack.getTag().equals(stack.getTag())) {
										hit = true;
										break;
									}
								}
							}

							if (!hit)
								return false;
						}

						return true;
					}

					@Override
					public void consumeCost() {
						if (experience > 0)
							capability.ply.giveExperienceLevels(-experience);

						for (var stack : resolved_stacks) {
							if (stack.getTag() == null) {
								for (var inv_stack : capability.ply.getInventory().items) {
									if (inv_stack.is(stack.getItem()) && inv_stack.getCount() >= stack.getCount()) {
										inv_stack.shrink(stack.getCount());
										break;
									}
								}
							} else {
								for (var inv_stack : capability.ply.getInventory().items) {
									if (inv_stack.is(stack.getItem()) && inv_stack.getCount() >= stack.getCount() && inv_stack.getTag() != null && inv_stack.getTag().equals(stack.getTag())) {
										inv_stack.shrink(stack.getCount());
										break;
									}
								}
							}
						}

						capability.ply.getInventory().setChanged();
					}
				};
			};
		}
	}
}
